-----------------------------------------------------------------------------------------------
-- Client Lua Script for DamageLocator
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"
 
-----------------------------------------------------------------------------------------------
-- DamageLocator Module Definition
-----------------------------------------------------------------------------------------------
local DamageLocator = {} 
 
-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
local ExistingIndicators = {}

-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function DamageLocator:new(Object)
    local Object = Object or {}
    setmetatable(Object, self)
    self.__index = self 

    -- initialize variables here

    return Object
end

function DamageLocator:Init()
    Apollo.RegisterAddon(self, false, "", {})
end
 

-----------------------------------------------------------------------------------------------
-- DamageLocator OnLoad
-----------------------------------------------------------------------------------------------
function DamageLocator:OnLoad()
    -- load our form file
	self.xmlDoc = XmlDoc.CreateFromFile("DamageLocator.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
	--Apollo.RegisterEventHandler("TargetedByUnit", "OnTargeted", self)
	--Apollo.RegisterEventHandler("UnitEnteredCombat", "OnCombat", self)
	Apollo.RegisterEventHandler("CombatLogDamage", "OnLogDamage", self)
end

-----------------------------------------------------------------------------------------------
-- DamageLocator OnDocLoaded
-----------------------------------------------------------------------------------------------
function DamageLocator:OnDocLoaded()

	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
	    self.wndMain = Apollo.LoadForm(self.xmlDoc, "DamageLocatorForm", nil, self)
		if self.wndMain == nil then
			Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
			return
		end
		
	    self.wndMain:Show(false, true)
		GeminiGUI = Apollo.GetPackage("Gemini:GUI-1.0").tPackage
		
		--TargetedByUnit
		
	end
end

-----------------------------------------------------------------------------------------------
-- DamageLocator Functions
-----------------------------------------------------------------------------------------------
-- Define general functions here

function DamageLocator:ValidateUnitIntegrity(UnitInstance)
	return UnitInstance ~= nil and UnitInstance.GetPosition ~= nil --The function can sometimes be nil. I don't know why.
end

function DamageLocator:CastTo2D(Vec3)
	return Vector2.New({Vec3.x, Vec3.z})
end

function DamageLocator:Sign(n)
	if n < 0 then
		return -1
	elseif n >= 0 then
		return 1
	end
end

function DamageLocator:Vec2ToAngle(Vec)
	return math.atan2(Vec.x, Vec.y)
end

function DamageLocator:AngleToVec2(Angle)
	return Vector2.New(math.cos(Angle), math.sin(Angle)):Normal() --This should already be normalized but I want to ensure it all.
end

function DamageLocator:GetPartialDirection(A, B)
	local ARot = DamageLocator:Vec2ToAngle(A)
	local BRot = DamageLocator:Vec2ToAngle(B)
	return DamageLocator:AngleToVec2((ARot - BRot) - math.pi/2)
end

function DamageLocator:DirectionToMe(TargetUnit)
	local Player = GameLib:GetPlayerUnit()
	if Player == nil then
		return Vector2.Zero()
	end
	local PTable0 = TargetUnit:GetPosition()
	local PTable1 = Player:GetPosition()
	local PFacing = Player:GetFacing()
	local Direction = DamageLocator:CastTo2D((Vector3.New(PTable1) - Vector3.New(PTable0))):Normal()
	local Facing = DamageLocator:CastTo2D(Vector3.New(PFacing)):Normal()
	return DamageLocator:GetPartialDirection(Direction, Facing)
end

function DamageLocator:CreateWindow(Indicator)
	local IndicatorObjectDefinition = {
		Name			= "Watcher" .. tostring(Indicator):sub(7),
		Template		= "Default",
		UseTemplateBG	= false,
		Picture			= true,
		Border			= false,
		AnchorCenter	= { 0, 0 }
	}
	
	local Window = GeminiGUI:Create(IndicatorObjectDefinition):GetInstance()
	return Window
end

function DamageLocator:MakeIndicator(TargetUnit)
	--Unit's messed up
	if not DamageLocator:ValidateUnitIntegrity(TargetUnit) then
		return
	end
	
	--Do NOT create multiple arrows for the same unit!
	--Test if it exists already.
	if ExistingIndicators[TargetUnit:GetId()] ~= nil then
		--It does exist. Call the revive function of the indicator and then we'll be good to go.
		local Indicator = ExistingIndicators[TargetUnit:GetId()]
		if Indicator.ThisId == TargetUnit:GetId() and (os.clock() - Indicator.CreationTime) < Indicator.LifeTime then --There's a chance that the id could be altered for am unknown reason.
			Indicator:Revive()
			return
		end --It prevents new arrows if one fades fully on a foe and they hit again, the condition should fix that.
		ExistingIndicators[TargetUnit:GetId()] = nil --Set it to nil too.
	end
	
	local Indicator = {}
	Indicator.Object = DamageLocator:CreateWindow(Indicator)	--Bless the man who wrote this library, the one to create windows in realtime.
	Indicator.Target = TargetUnit 								--Register the target unit
	Indicator.ThisId = TargetUnit:GetId()						--Register the ID of this target
	Indicator.LifeTime = 5										--Set the lifetime of the arrow.
	Indicator.TimerName = "Indicator" .. tostring(Indicator)	--Create a unique ID for this indicator's timer by getting the table's tag via tostring()
	Indicator.CreationTime = os.clock()							--Set the creation time
	Indicator.Stop = false
	
	--Append this indicator to the registry so that no redundant indicators are made.
	ExistingIndicators[Indicator.ThisId] = Indicator
	
	--Update the sprite for this indicator to a white arrow sprite and show it, then set its color to a red color.
	Indicator.Object:SetSprite("ClientSprites:MiniMapPlayerArrow")
	Indicator.Object:Show(true, true)
	Indicator.Object:SetBGColor(CColor.new(1, 0.2, 0))
	
	--Move the arrow to a position on-screen.
	function Indicator:MoveTo(X, Y)
		local MX = X * 185
		local MY = Y * 185
		Indicator.Object:SetAnchorPoints(0.5, 0.5, 0.5, 0.5)
		Indicator.Object:SetAnchorOffsets(-25 + MX, -25 - MY, 25 + MX, 25 - MY)
	end
	
	--Update the indicator. This is caleld via the timers created below.
	function Indicator:Update()
		if Indicator.Stop then
			return
		end
		
		local LifetimePassed = (os.clock() - Indicator.CreationTime) / Indicator.LifeTime
		local ShouldStop = LifetimePassed >= 1 or Indicator.Target == nil or Indicator.Target:IsValid() == false or Indicator.Target:IsDead() or not Indicator.Target:IsInCombat()
		if ShouldStop then
			--If the indicator has timed out or the target has been destroyed before it times out, remove it.
			--Remove the indicator before this call is finished so that the revive function can get safely called before removal, if we get a situation where said timing occurs.
			
			--This is where a bad error occurs due to what we're using to index it being nil.
			--Luckily, the very cause of that error (index is nil) also means that the lua engine will have cleared it out for us, so it's ok if we miss it because of that.
			if Indicator.ThisID ~= nil then
				ExistingIndicators[Indicator.ThisID] = nil
			end
			
			Apollo.StopTimer(Indicator.TimerName)
			Indicator.Object:Destroy()
			Indicator.Stop = true
			return
		end
		Indicator.Object:SetBGColor(CColor.new(1, 0.2, 0, 1 - LifetimePassed))
		local Facing = DamageLocator:DirectionToMe(Indicator.Target)
		Indicator.Object:SetRotation(math.deg(DamageLocator:Vec2ToAngle(Facing)))
		Indicator:MoveTo(Facing.x, Facing.y)
	end
	
	--Revive the indicator. This will set the existing indicator's creation time to the current time, effectively resetting it.
	function Indicator:Revive()
		Indicator.CreationTime = os.clock()
	end
	
	Apollo.RegisterTimerHandler(Indicator.TimerName, "Update", Indicator)
	Apollo.CreateTimer(Indicator.TimerName, 0.016, true)
	
	return Indicator
end

function DamageLocator:OnLogDamage(InfoTable)	
	local Target = InfoTable.unitTargetOwner or InfoTable.unitTarget
	if Target and Target:IsThePlayer() then
		DamageLocator:MakeIndicator(InfoTable.unitCaster)
		DamageLocator:MakeIndicator(InfoTable.unitCasterOwner)
	end
end

-----------------------------------------------------------------------------------------------
-- DamageLocator Instance
-----------------------------------------------------------------------------------------------
local DamageLocatorInst = DamageLocator:new()
DamageLocatorInst:Init()
